import hashlib

if 'ripemd160' not in hashlib.algorithms_available:
	from ripemd import ripemd160 as ripemd160_impl


def _factory():
	return hashlib.new('ripemd160') if 'ripemd160' in hashlib.algorithms_available else ripemd160_impl.new()


def ripemd160(data):
	"""Calculates RIPEMD-160 hash of data."""

	builder = _factory()
	builder.update(data)
	return builder.digest()



import yaml

from .AccountDescriptorRepository import AccountDescriptorRepository
from .NodeDescriptorRepository import NodeDescriptorRepository


class BlockchainSettings:
	"""Settings describing a blockchain."""

	def __init__(self, settings_dict):
		"""Creates blockchain settings from a dictionary."""
		self.blockchain = settings_dict['blockchain']
		self.network = settings_dict['network']
		self.nodes = NodeDescriptorRepository(settings_dict['nodes'])
		self.accounts = AccountDescriptorRepository(settings_dict['accounts'])

	@staticmethod
	def load_from_yaml(yaml_input):
		"""Loads settings from YAML."""
		settings_yaml = yaml.load(yaml_input, Loader=yaml.SafeLoader)
		return BlockchainSettings(settings_yaml)



import hashlib

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.hkdf import HKDF

from .CryptoTypes import SharedKey256
from .external import ed25519


class SharedKey:
	@staticmethod
	def _derive_shared_key(other_public_key_bytes, private_key_bytes, info, hashing_algorithm=hashlib.sha512):
		shared_secret = ed25519.derive_shared_secret_unsafe(other_public_key_bytes, private_key_bytes, hashing_algorithm)
		salt = bytes(32)
		hkdf = HKDF(algorithm=hashes.SHA256(), length=32, salt=salt, info=info)
		return SharedKey256(hkdf.derive(shared_secret))



import functools
import inspect
import operator
from enum import Enum, Flag

from .BaseValue import BaseValue
from .ByteArray import ByteArray
from .TransactionDescriptorProcessor import TransactionDescriptorProcessor


def _name_to_enum_value(mapping, enum_type, enum_value_name):
	if enum_value_name not in mapping:
		raise ValueError(f'unknown value {enum_value_name} for type {enum_type}')

	return mapping[enum_value_name]


def _build_type_hints_map(struct_value):
	type_hints = {}
	for key, hint in struct_value.TYPE_HINTS.items():
		rule_name = None
		if hint.startswith('array['):
			rule_name = hint
		elif hint.startswith('enum:'):
			rule_name = hint[len('enum:'):]
		elif hint.startswith('pod:'):
			rule_name = hint[len('pod:'):]
		elif hint.startswith('struct:'):
			rule_name = hint

		if rule_name:
			type_hints[key] = rule_name

	return type_hints


def _type_converter_factory(module, custom_type_converter, value):
	if custom_type_converter and custom_type_converter(value):
		return custom_type_converter(value)

	if isinstance(value, ByteArray):
		return getattr(module, type(value).__name__)(value.bytes)

	return value


class RuleBasedTransactionFactory:
	"""Rule based transaction factory."""

	def __init__(self, module, type_converter=None, type_rule_overrides=None):
		"""Creates a rule based transaction factory for use with catbuffer generated code."""
		self.module = module
		self.type_converter = lambda value: _type_converter_factory(self.module, type_converter, value)
		self.type_rule_overrides = type_rule_overrides or {}
		self.rules = {}

	def _get_module_class(self, name):
		return getattr(self.module, name)

	def add_pod_parser(self, name, pod_class):
		"""Creates wrapper for SDK POD types."""

		if pod_class in self.type_rule_overrides:
			# if an override has been set for the type, use it
			self.rules[name] = self.type_rule_overrides[pod_class]
			return

		self.rules[name] = lambda value: pod_class(value) if not isinstance(value, pod_class) else value

	def add_flags_parser(self, name):
		"""Creates flag type parser."""
		flags_class = self._get_module_class(name)
		string_to_enum = dict(map(lambda key: (key.name.lower(), key), flags_class))

		def parser(flags):
			if isinstance(flags, str):
				enum_array = list(map(lambda flag_name: _name_to_enum_value(string_to_enum, name, flag_name), flags.split(' ')))
				return functools.reduce(operator.or_, enum_array)

			if isinstance(flags, int):
				return flags_class(flags)

			return flags

		self.rules[name] = parser

	def add_enum_parser(self, name):
		"""Creates enum type parser."""
		enum_class = self._get_module_class(name)
		string_to_enum = dict(map(lambda key: (key.name.lower(), key), enum_class))

		def parser(enum_value):
			if isinstance(enum_value, str):
				return _name_to_enum_value(string_to_enum, name, enum_value)

			if isinstance(enum_value, int):
				return enum_class(enum_value)

			return enum_value

		self.rules[name] = parser

	def add_struct_parser(self, name):
		"""Creates struct parser (to allow nested parsing)."""
		struct_class = self._get_module_class(name)

		def parser(struct_descriptor):
			struct_processor = self._create_processor(struct_descriptor)
			struct_value = struct_class()

			all_type_hints = _build_type_hints_map(struct_value)
			struct_processor.set_type_hints(all_type_hints)

			struct_processor.copy_to(struct_value)
			return struct_value

		self.rules[f'struct:{name}'] = parser

	def add_array_parser(self, name):
		"""Creates array type parser, based on some existing element type parser."""
		element_rule = self.rules[name]
		element_name = name[len('struct:'):] if name.startswith('struct:') else name

		def parser(values):
			return list(map(element_rule, values))

		self.rules[f'array[{element_name}]'] = parser

	def autodetect(self):
		"""Autodetects rules using reflection."""
		for class_name in dir(self.module):
			cls = getattr(self.module, class_name)
			if not inspect.isclass(cls):
				continue

			if issubclass(cls, BaseValue) and cls != BaseValue:
				self.add_pod_parser(class_name, cls)

			if issubclass(cls, Enum) and cls != Enum and cls != Flag:
				if issubclass(cls, Flag):
					self.add_flags_parser(class_name)
				else:
					self.add_enum_parser(class_name)

	def create_from_factory(self, factory, descriptor):
		"""Creates an entity from a descriptor using a factory."""
		processor = self._create_processor(descriptor)
		entity_type = processor.lookup_value('type')
		entity = factory(entity_type)

		all_type_hints = _build_type_hints_map(entity)
		processor.set_type_hints(all_type_hints)
		processor.copy_to(entity, ['type'])

		self._auto_encode_strings(entity)
		return entity

	def _create_processor(self, descriptor):
		return TransactionDescriptorProcessor(descriptor, self.rules, self.type_converter)

	@staticmethod
	def _auto_encode_strings(entity):
		for key, value in vars(entity).items():
			if isinstance(value, str):
				setattr(entity, key, value.encode('utf8'))



import cryptography

from symbolchain.CryptoTypes import PublicKey
from symbolchain.impl.CipherHelpers import decode_aes_cbc, decode_aes_gcm, encode_aes_cbc, encode_aes_gcm
from symbolchain.nc import Message, MessageType
from symbolchain.nem.KeyPair import KeyPair
from symbolchain.nem.SharedKey import SharedKey


class MessageEncoder:
	"""Encrypts and encodes messages between two parties."""

	def __init__(self, key_pair: KeyPair):
		"""Creates message encoder around key pair."""
		self.key_pair = key_pair

	def try_decode(self, recipient_public_key, encoded_message: Message):
		"""Tries to decode encoded message, returns tuple:
		* True, message - if message has been decoded and decrypted
		* False, encoded_message - otherwise
		"""

		if MessageType.ENCRYPTED != encoded_message.message_type:
			raise RuntimeError('invalid message format')

		try:
			message = decode_aes_gcm(SharedKey, self.key_pair, recipient_public_key, encoded_message.message)
			return True, message
		except cryptography.exceptions.InvalidTag:
			pass

		try:
			message = decode_aes_cbc(SharedKey, self.key_pair, recipient_public_key, encoded_message.message)
			return True, message
		except ValueError as exception:
			exceptions = [
				'Invalid padding bytes',
				'Invalid IV size',
				'The length of the provided data is not a multiple of the block length']
			if not any(map(lambda message: message in str(exception), exceptions)):
				raise

		return False, encoded_message

	def encode_deprecated(self, recipient_public_key: PublicKey, message: bytes):
		"""Encodes message to recipient using deprecated encryption and key derivation."""

		salt, initialization_vector, cipher_text = encode_aes_cbc(SharedKey, self.key_pair, recipient_public_key, message)

		encoded_messsage = Message()
		encoded_messsage.message_type = MessageType.ENCRYPTED
		encoded_messsage.message = salt + initialization_vector + cipher_text
		return encoded_messsage

	def encode(self, recipient_public_key: PublicKey, message: bytes):
		"""Encodes message to recipient using recommended format."""

		tag, initialization_vector, cipher_text = encode_aes_gcm(SharedKey, self.key_pair, recipient_public_key, message)

		encoded_messsage = Message()
		encoded_messsage.message_type = MessageType.ENCRYPTED
		encoded_messsage.message = tag + initialization_vector + cipher_text
		return encoded_messsage



import hashlib
import hmac
import secrets

from mnemonic import Mnemonic

from .BufferWriter import BufferWriter
from .CryptoTypes import PrivateKey


class Bip32Node:
	"""Representation of a BIP32 node."""

	def __init__(self, hmac_key, data):
		"""Creates a BIP32 node around a key and data."""
		hmac_result = hmac.new(hmac_key, data, hashlib.sha512).digest()

		self.private_key = PrivateKey(hmac_result[0:PrivateKey.SIZE])
		self.chain_code = hmac_result[PrivateKey.SIZE:]

	def derive_one(self, identifier):
		"""Derives a direct child node with specified identifier."""
		hmac_data_writer = BufferWriter('big')
		hmac_data_writer.write_int(0, 1)
		hmac_data_writer.write_bytes(self.private_key.bytes)
		hmac_data_writer.write_int(0x80000000 | identifier, 4)
		return Bip32Node(self.chain_code, hmac_data_writer.buffer)

	def derive_path(self, path):
		"""Derives a descendent node with specified path."""
		next_node = self
		for identifier in path:
			next_node = next_node.derive_one(identifier)

		return next_node


class Bip32:
	"""Factory of BIP32 root nodes """

	def __init__(self, curve_name='ed25519', mnemonic_language='english'):
		"""Creates a BIP32 root node factory."""
		self.root_hmac_key = (curve_name + ' seed').encode('utf8')
		self.mnemonic_language = mnemonic_language

	def from_seed(self, seed):
		"""Creates a BIP32 root node from a seed."""
		return Bip32Node(self.root_hmac_key, seed)

	def from_mnemonic(self, mnemonic, password):
		"""Creates a BIP32 root node from a BIP39 mnemonic and password."""
		return self.from_seed(Mnemonic(self.mnemonic_language).to_seed(mnemonic, password))

	def random(self, seed_length=32):
		"""Creates a random BIP32 mnemonic."""
		return Mnemonic(self.mnemonic_language).to_mnemonic(secrets.token_bytes(seed_length))



import secrets

from .ByteArray import ByteArray


class Hash256(ByteArray):
	"""Represents a 256-bit hash."""

	SIZE = 32

	def __init__(self, hash256):
		"""Creates a hash from bytes or a hex string."""
		super().__init__(self.SIZE, hash256, Hash256)

	def __repr__(self):
		return f'Hash256(\'{str(self)}\')'

	@staticmethod
	def zero():
		"""Creates a zeroed hash."""
		return Hash256(bytes([0] * Hash256.SIZE))


class PrivateKey(ByteArray):
	"""Represents a private key."""

	SIZE = 32

	def __init__(self, private_key):
		"""Creates a private key from bytes or a hex string."""
		super().__init__(self.SIZE, private_key, PrivateKey)

	def __repr__(self):
		return f'PrivateKey(\'{str(self)}\')'

	@staticmethod
	def random():
		"""Generates a random private key."""
		return PrivateKey(secrets.token_bytes(PrivateKey.SIZE))


class PublicKey(ByteArray):
	"""Represents a public key."""

	SIZE = 32

	def __init__(self, public_key):
		"""Creates a public key from bytes or a hex string."""
		super().__init__(self.SIZE, public_key.bytes if isinstance(public_key, PublicKey) else public_key, PublicKey)

	def __repr__(self):
		return f'PublicKey(\'{str(self)}\')'


class SharedKey256(ByteArray):
	"""Represents 256-bit symmetric encryption key."""

	SIZE = 32

	def __init__(self, key):
		"""Creates a key from bytes or a hex string."""
		super().__init__(self.SIZE, key, SharedKey256)

	def __repr__(self):
		return f'SharedKey256(\'{str(self)}\')'


class Signature(ByteArray):
	"""Represents a signature."""

	SIZE = 64

	def __init__(self, signature):
		"""Creates a signature from bytes or a hex string."""
		super().__init__(self.SIZE, signature, Signature)

	def __repr__(self):
		return f'Signature(\'{str(self)}\')'

	@staticmethod
	def zero():
		"""Creates a zeroed signature."""
		return Signature(bytes([0] * Signature.SIZE))



import sha3

from symbolchain.ripemd160 import ripemd160


def ripemd_keccak_256(payload):
	"""Hashes payload with keccak 256 and then hashes the result with ripemd160."""

	return ripemd160(sha3.keccak_256(payload).digest())



import operator
from abc import ABC, abstractmethod


class Ordered(ABC):
	@abstractmethod
	def _cmp(self, other, operation):
		pass

	def __lt__(self, other):
		return self._cmp(other, operator.lt)

	def __le__(self, other):
		return self._cmp(other, operator.le)

	def __gt__(self, other):
		return self._cmp(other, operator.gt)

	def __ge__(self, other):
		return self._cmp(other, operator.ge)



import hashlib
import math

from mnemonic import Mnemonic


class DiceMnemonicGenerator:
	"""Generates BIP39 mnemonics from dice rolls."""

	def __init__(self, num_die_sides=6, mnemonic_language='english'):
		"""Creates a generator."""
		self.num_die_sides = num_die_sides
		self.mnemonic_language = mnemonic_language
		self.rolls = []

	def add_roll(self, value):
		"""Adds a roll."""
		if 1 <= value <= self.num_die_sides:
			self.rolls.append(value)
			return

		raise ValueError(f'roll value must be between 1 and {self.num_die_sides}, inclusive')

	def frequencies(self):
		"""Gets the number of times each number was rolled."""
		counts = [0] * self.num_die_sides
		for roll in self.rolls:
			counts[roll - 1] += 1

		return counts

	def to_mnemonic(self, shrink_wrap=False):
		"""Generates a mnemonic by hashing the roll input. Disabling `shrink_wrap` is the same method used by coldcard wallet."""
		seed = hashlib.sha256(''.join(map(str, self.rolls)).encode('utf8')).digest()
		if not shrink_wrap:
			return (self._seed_to_mnemonic(seed), min(256, len(self.rolls) * -math.log2(1 / self.num_die_sides)))

		seed_length = self._calculate_shrink_wrap_seed_length()
		return (self._seed_to_mnemonic(seed[0:seed_length]), 8 * seed_length)

	def _calculate_shrink_wrap_seed_length(self):
		min_seed_length = 16
		max_seed_length = 32

		total_roll_possibilities = self.num_die_sides ** len(self.rolls)
		if total_roll_possibilities < 2 ** (8 * min_seed_length):
			raise ValueError(f'shrink_wrap requires at least 128 bits of entropy but only {total_roll_possibilities} bits provided')

		seed_length = min_seed_length
		while 2 ** (8 * (seed_length + 4)) < total_roll_possibilities:
			seed_length += 4

		return min(max_seed_length, seed_length)

	def _seed_to_mnemonic(self, seed):
		return Mnemonic(self.mnemonic_language).to_mnemonic(seed)



import sha3
from nacl.bindings import (
	crypto_core_ed25519_is_valid_point,
	crypto_core_ed25519_scalar_add,
	crypto_core_ed25519_scalar_mul,
	crypto_core_ed25519_scalar_reduce,
	crypto_core_ed25519_sub,
	crypto_scalarmult_ed25519_base,
	crypto_scalarmult_ed25519_base_noclamp,
	crypto_scalarmult_ed25519_noclamp
)

from ..CryptoTypes import PrivateKey, PublicKey, Signature


def _generate_nonce(secret_key, message):
	hashed_secret = sha3.keccak_512(secret_key).digest()
	hashobj = sha3.keccak_512()
	hashobj.update(hashed_secret[32:])
	hashobj.update(message)
	final_hash = hashobj.digest()
	return final_hash


def _is_reduced_s(encoded_s):
	reduced = crypto_core_ed25519_scalar_reduce(encoded_s + bytes(32))
	return reduced == encoded_s


def _is_canonical_s(encoded_s):
	# reject zero S
	if encoded_s == bytes(32):
		return False

	# reject non-reduced S
	return _is_reduced_s(encoded_s)


class KeyPair:
	"""Represents an ED25519 private and public key."""

	def __init__(self, private_key):
		"""Creates a key pair from a private key."""
		self._sk = private_key.bytes[::-1]

		# nacl does clamping
		hashed_secret = sha3.keccak_512(self._sk).digest()
		self._pk = crypto_scalarmult_ed25519_base(hashed_secret[:32])

	@property
	def public_key(self):
		"""Gets the public key."""
		return PublicKey(self._pk)

	@property
	def private_key(self):
		"""Gets the private key."""
		return PrivateKey(self._sk[::-1])

	def sign(self, message):
		"""Signs a message with the private key."""

		# pylint: disable=invalid-name
		# r = H(privHash[256:512] || data)
		r = _generate_nonce(self._sk, message)

		# R = r * base point
		r = crypto_core_ed25519_scalar_reduce(r)
		R = crypto_scalarmult_ed25519_base_noclamp(r)

		# h = H(encodedR || public || data)
		hashobj = sha3.keccak_512()
		hashobj.update(R[:32])
		hashobj.update(self._pk)
		hashobj.update(message)
		h = hashobj.digest()
		h = crypto_core_ed25519_scalar_reduce(h)

		# hash and clamp private key
		a = bytearray(sha3.keccak_512(self._sk).digest()[:32])
		a[0] &= 0xF8
		a[31] &= 0x7F
		a[31] |= 0x40
		a = bytes(a)

		# S = (r + h * a) mod L
		h_a = crypto_core_ed25519_scalar_mul(a, h)
		S = crypto_core_ed25519_scalar_add(r, h_a)

		# ensure resulting signature is canonical (this is a sanity check that should never get triggered)
		assert bytes(32) == S or _is_reduced_s(S)

		act_sig = R + S
		return Signature(act_sig)


class Verifier:
	"""Verifies signatures signed by a single key pair."""

	def __init__(self, public_key):
		"""Creates a verifier from a public key."""
		if bytes(32) == public_key.bytes:
			raise ValueError('public key cannot be zero')

		self._pk = public_key.bytes

	def verify(self, message, signature):
		"""Verifies a message signature."""

		encoded_r = signature.bytes[:32]
		encoded_s = signature.bytes[32:]

		if not _is_canonical_s(encoded_s):
			return False

		# h = H(encodedR || public || data)
		hashobj = sha3.keccak_512()
		hashobj.update(encoded_r)
		hashobj.update(self._pk)
		hashobj.update(message)
		h = crypto_core_ed25519_scalar_reduce(hashobj.digest())  # pylint: disable=invalid-name

		# note: this check covers following:
		#  * (y) coord is canonical (< 2^255 - 19)
		#  * has small order - this is slightly stronger check,
		#    we don't do that in the client as we verify sigs with pub keys that are known to belong to accounts
		#  * point is on curve
		#  * point is in main subgroup
		if not crypto_core_ed25519_is_valid_point(self._pk):
			return False

		# R = encodedS * B - h * A
		s_b = crypto_scalarmult_ed25519_base_noclamp(encoded_s)
		h_a = crypto_scalarmult_ed25519_noclamp(h, self._pk)

		computed_r = crypto_core_ed25519_sub(s_b, h_a)
		return computed_r == encoded_r



from binascii import hexlify


class BufferReader:
	"""Reads data from an in memory buffer."""

	def __init__(self, buffer, byte_order='little'):
		"""Creates a reader with specified byte order."""
		self.buffer = buffer
		self.byte_order = byte_order
		self.offset = 0

	@property
	def eof(self):
		"""Returns true if the reader is at eof."""
		return self.offset == len(self.buffer)

	def read_int(self, count):
		"""Reads an integer."""
		return int.from_bytes(self.read_bytes(count), self.byte_order)

	def read_string(self, count):
		"""Reads a string."""
		return self.read_bytes(count).decode('utf8')

	def read_hex_string(self, count):
		"""Reads a hex string."""
		return hexlify(self.read_bytes(count)).decode('utf8').upper()

	def read_bytes(self, count):
		"""Reads bytes."""
		buffer = self.buffer[self.offset:self.offset + count]
		self.offset += count
		return buffer
